#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <windows.h>
#include <GL/gl.h>

#include "resource.h"

#define MAXITERATIONS 100

#define COLOR(r,g,b) (((r)<<16)|((g)<<8)|(b))
#define getR(x) (x>>16)
#define getG(x) ((x>>8)&255)
#define getB(x) (x&255)

//TODO: 
// *Add quality slider to adjust pixel size
// *Add text to let user know the program is rendering

int DrawArea[400][700];

int Colors[] = 
{
	COLOR(127,0,0),
	COLOR(255,0,0),
	COLOR(255,127,0),
	COLOR(255,255,0),
	COLOR(127,255,0),
	COLOR(0,255,0),
	COLOR(0,255,127),
	COLOR(0,255,255),
	COLOR(0,127,255),
	COLOR(0,0,255),
	COLOR(255,0,255),
};

HDC hDC;
HGLRC hRC;

LRESULT CALLBACK GLProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
INT_PTR CALLBACK DlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
void initGL(HWND hwnd);
void PutPixel(double x, double y);
void DrawFractal();
void Reset();

HWND hwndWindow;
HWND hwndChild;

char MouseDown;

double xStart, yStart;
double xScale, yScale;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	WNDCLASSEX wndCls;
	MSG msg;

	ZeroMemory(&wndCls, sizeof(WNDCLASSEX));
	wndCls.cbSize = sizeof(WNDCLASSEX);
	wndCls.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	wndCls.hInstance = hInstance;
	wndCls.lpfnWndProc = GLProc;
	wndCls.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndCls.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndCls.hbrBackground = CreateSolidBrush(0);
	wndCls.lpszClassName = "Windy";
	wndCls.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

	RegisterClassEx(&wndCls);
	
	hwndWindow = CreateDialog(
		hInstance,
		MAKEINTRESOURCE(IDD_DIALOG1),
		NULL, DlgProc);
		
	hwndChild = CreateWindowEx(
		0, "Windy", "GL", WS_CHILD /*| WS_BORDER*/,
		41, 10, 700, 400,
		hwndWindow, NULL, hInstance, NULL);
	
	ShowWindow(hwndChild, 1);
	UpdateWindow(hwndChild);
	
	initGL(hwndChild);
	Reset();
	
	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return msg.wParam;
}

void initGL(HWND hwnd)
{
	hDC = GetDC(hwnd);	

	PIXELFORMATDESCRIPTOR pfd=
	{
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW |
		PFD_SUPPORT_OPENGL |
		PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		16,0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0,
		16, 0, 0, PFD_MAIN_PLANE,
		0, 0, 0, 0
	};

	unsigned int PixelFormat = ChoosePixelFormat(hDC,&pfd);
	SetPixelFormat(hDC,PixelFormat,&pfd);

	hRC = wglCreateContext(hDC);

	wglMakeCurrent(hDC,hRC);

	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,700,400,0,0,1);
	glViewport(0,0,700,400);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(255,255,255,255);
	glColor3f(0,0,0);
	
	glHint(GL_POLYGON_SMOOTH, GL_NICEST);
	glEnable(GL_POLYGON_SMOOTH);

	glClear(GL_COLOR_BUFFER_BIT);
	SwapBuffers(hDC);
}

void PutPixel(double x, double y)
{
	glVertex2f(x, y);
	glVertex2f(x, y + 1);
	glVertex2f(x + 1, y + 1);
	glVertex2f(x + 1, y);
}

void MakeImage()
{
	int x,y,color;
	
	glClear(GL_COLOR_BUFFER_BIT);
	glBegin(GL_QUADS);
	
	for(y = 0; y < 400; y++)		
		for(x = 0; x < 700; x++)
		{
			color = DrawArea[y][x];
				
			glColor3f(
			getR(color)/255.0,
			getG(color)/255.0,
			getB(color)/255.0);
				
			PutPixel(x, y);
		}
	
	glEnd();
	SwapBuffers(hDC);
}

void HandleMouse(int x, int y)
{	
	xStart = (xStart + (x * xScale)) - (700 * xScale) / 4.0;
	yStart = (yStart - (y * yScale)) + (400 * yScale) / 4.0;
	
	xScale /= 2;
	yScale /= 2;
	
	DrawFractal();
}

void Reset()
{
	xStart = -2.5;
	yStart = 1;
	xScale = yScale = 0.005;
	
	DrawFractal();
}

void DrawFractal()
{
	int ScreenX, ScreenY;
	double x0, y0, xn, yn, xTemp;
	int count;
	
	for(ScreenY = 0; ScreenY < 400; ScreenY++)
	{
		for(ScreenX = 0; ScreenX < 700; ScreenX++)
		{
			x0 = xStart + (ScreenX * xScale);
			y0 = yStart - (ScreenY * yScale);
			
			xn = 0;
			yn = 0;
			
			count = 0;
			
			while(xn*xn + yn*yn <= (2*2) && count < MAXITERATIONS)
			{
				xTemp = xn*xn - yn*yn + x0;
				yn = 2*xn*yn + y0;
				
				xn = xTemp;
				
				count++;
			}
			
			if(count == MAXITERATIONS)
				DrawArea[ScreenY][ScreenX] = COLOR(0,0,0);
			else
				DrawArea[ScreenY][ScreenX] = Colors[count % 11];
		}
	}
	
	MakeImage();
}

INT_PTR CALLBACK DlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{	
	switch(message)
	{			
		case WM_DESTROY:
		case WM_CLOSE:
			ReleaseDC(WindowFromDC(hDC), hDC);
			wglDeleteContext(hRC);
			PostQuitMessage(0);
			return 1;
			
		case WM_COMMAND:
			switch(wParam & 0xFF)
			{
				case RESET:
					Reset();
					break;
			}
			return 1;
	}
	return 0;
}

LRESULT CALLBACK GLProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{	
	switch(message)
	{
		case WM_LBUTTONDOWN:
			HandleMouse(lParam & 0xFFFF, (lParam >> 16) & 0xFFFF);
			return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}
